#include <stdio.h>
#include <stdlib.h>
#include <sys/socket.h>
#include <net/if.h>
#include <uci.h>
#include <string.h>
#include <pthread.h>

#include <netlink/netlink.h>
#include <netlink/route/rtnl.h>
#include <netlink/socket.h>
#include <netlink/msg.h>

#include <linux/if_bridge.h>

#include <sys/ioctl.h>
#include <net/if.h>

#include <libubus.h>

#include <easy/easy.h>

#include <map2.h>
#include <i1905_wsc.h>
#include "config.h"
#include "agent.h"

#define dbg(...) fprintf(stderr, __VA_ARGS__)
#define warn(...) fprintf(stderr, __VA_ARGS__)

#define TS_IFACE_MAX_NUM (2*WIFI_IFACE_MAX_NUM)
#define MAX_VIDS (WIFI_IFACE_MAX_NUM - 4)

static struct nl_sock *nl_main_sk;
static struct nl_sock *nl_bridge_sk;
static struct nl_cb *nl_bridge_cb;

static bool ts_bridge_dump_done;
static bool ts_check_vlan_settings;
bool ts_need_reconfig;

struct ts_iface {
	char name[IFNAMSIZ];
	uint16_t vid;
	int ifi_index;
	struct bridge_vlan_info cur_vinfo[MAX_VIDS];
	bool is_bsta;
};

static struct ts_iface ts_iface_array[TS_IFACE_MAX_NUM];
static uint16_t ts_primary_vid;

static uint16_t ts_all_vids[MAX_VIDS];
static unsigned int ts_num_vids = 0;

static bool ts_is_our_vid(uint16_t vid)
{
	int i;

	for (i = 0; i < ts_num_vids; i++) {
		if (ts_all_vids[i] == vid)
			return true;
	}

	return false;
}

static struct ts_iface *ts_find_iface(const char *ifname)
{
	int i;

	for (i = 0; i < TS_IFACE_MAX_NUM; i++) {
		if (strncmp(ts_iface_array[i].name, ifname , 16) == 0)
			return &ts_iface_array[i];
	}

	return NULL;
}

static struct ts_iface *ts_configure_iface(const char *ifname, uint16_t vid)
{
	struct ts_iface *tsif;
	int i;

	for (i = 0; i < TS_IFACE_MAX_NUM; i++) {
		tsif = &ts_iface_array[i];
		if (tsif->name[0] == '\0') {
			strncpy(tsif->name, ifname, IFNAMSIZ);
			ts_check_vlan_settings = true;
			break;
		}

		if (strncmp(tsif->name, ifname, IFNAMSIZ) == 0)
			break;
	}

	if (i >= TS_IFACE_MAX_NUM) {
		warn("Number of TS interfaces too small\n");
		return NULL;
	}

	if (tsif->vid != vid) {
		tsif->vid = vid;
		ts_check_vlan_settings = true;
	}

	if (vid != 0 && !ts_is_our_vid(tsif->vid)) {
		if (ts_num_vids < MAX_VIDS)
			ts_all_vids[ts_num_vids++] = tsif->vid;
		else
			warn("Too many vids\n");
	}

	return tsif;
}

static void tsif_check_attr(struct ts_iface *tsif, struct rtattr *attr)
{
	struct bridge_vlan_info *vinfo;
	struct rtattr *cur;
	int rem = RTA_PAYLOAD(attr);
	int i;

	for (cur = RTA_DATA(attr); RTA_OK(cur, rem); cur = RTA_NEXT(cur, rem)) {
		if (cur->rta_type != IFLA_BRIDGE_VLAN_INFO)
			continue;

		vinfo = RTA_DATA(cur);

		dbg("%s %s vid %d flags %0x\n", __func__, tsif->name, vinfo->vid, vinfo->flags);

		for (i = 0; i < MAX_VIDS; i++) {
			if (tsif->cur_vinfo[i].vid == 0) {
				tsif->cur_vinfo[i] = *vinfo;
				break;
			}
		}

		if (i >= MAX_VIDS)
			warn("%s cur_vinfo array too small\n", tsif->name);
	}
}

static int ts_nl_get_cb(struct nl_msg *msg, void *arg)
{
	struct nlmsghdr *nh = nlmsg_hdr(msg);
	struct ifinfomsg *ifi = NLMSG_DATA(nh);
	struct rtattr *attr;
	struct ts_iface *tsif;
	char name[16];
	char *ifname;
	int rem;

	if (!nh || !ifi)
		return NL_SKIP;

	dbg("%s family %u index %u\n", __func__, ifi->ifi_family, ifi->ifi_index);

	if (nh->nlmsg_type != RTM_NEWLINK)
		return NL_SKIP;

	if (ifi->ifi_family != AF_BRIDGE)
		return NL_SKIP;

	ifname = if_indextoname(ifi->ifi_index, name);
	if (!ifname) {
		ts_bridge_dump_done = true;
		return NL_STOP;
	}

	tsif = ts_find_iface(ifname);
	if (!tsif)
		return NL_SKIP;

	attr = IFLA_RTA(ifi);
	rem = nh->nlmsg_len - NLMSG_LENGTH(sizeof(*ifi));
	while (RTA_OK(attr, rem)) {
		if (attr->rta_type == IFLA_AF_SPEC)
			tsif_check_attr(tsif, attr);

		attr = RTA_NEXT(attr, rem);
	}

	tsif->ifi_index = ifi->ifi_index;

	return NL_SKIP;
}

static int ts_nl_finish_cb(struct nl_msg *msg, void *arg)
{
	dbg("%s\n", __func__);

	ts_bridge_dump_done = true;

	return NL_STOP;
}

static int ts_nl_error_cb(struct sockaddr_nl *nla, struct nlmsgerr *err, void *arg)
{
	/* TODO print error */
	dbg("%s\n", __func__);

	ts_bridge_dump_done = true;

	return NL_STOP;
}

void ts_read_bridge_vlan(const char *br_name)
{
	static struct ifinfomsg ifi = {
		.ifi_family = AF_BRIDGE
	};
	static struct rtattr ext_req = {
		.rta_type = IFLA_EXT_MASK,
		.rta_len = RTA_LENGTH(sizeof(uint32_t)),
	};
	uint32_t filter = RTEXT_FILTER_BRVLAN;
	struct nl_msg *nlmsg;
	int ret;

	/* TODO check if this works when there are more than one bridge */
	ifi.ifi_index = if_nametoindex(br_name);
	if (ifi.ifi_index <= 0)
		return;


	nlmsg = nlmsg_alloc_simple(RTM_GETLINK, NLM_F_DUMP);
	if (!nlmsg)
		return;

	dbg("%s START\n", __func__);

	ret = nlmsg_append(nlmsg, &ifi, sizeof(ifi), 0);
	ret |= nlmsg_append(nlmsg, &ext_req, sizeof(ext_req), NLMSG_ALIGNTO);
	ret |= nlmsg_append(nlmsg, &filter, sizeof(filter), 0);

	if (ret)
		goto free;

	ts_bridge_dump_done = false;

	if (nl_send_auto_complete(nl_bridge_sk, nlmsg) < 0)
		ts_bridge_dump_done = true;

	while (!ts_bridge_dump_done)
		nl_recvmsgs(nl_bridge_sk, nl_bridge_cb);
free:
	nlmsg_free(nlmsg);

	dbg("%s END\n", __func__);
}

static int send_bridge_vlan_info(int nlmsg_type, struct ifinfomsg *ifi, struct bridge_vlan_info *vinfo, int n)
{
	struct nl_msg *nlmsg;
	struct nlattr *attr;
	int ret = -1;
	int i;

	nlmsg = nlmsg_alloc_simple(nlmsg_type, NLM_F_REQUEST);
	if (!nlmsg)
		return -1;

	nlmsg_append(nlmsg, ifi, sizeof(*ifi), 0);

	attr = nla_nest_start(nlmsg, IFLA_AF_SPEC);
	if (!attr)
		goto err;

	for (i = 0; i < n; i++)
		nla_put(nlmsg, IFLA_BRIDGE_VLAN_INFO, sizeof(vinfo[i]), &vinfo[i]);

	nla_nest_end(nlmsg, attr);

	ret = nl_send_auto_complete(nl_main_sk, nlmsg);

err:
	nlmsg_free(nlmsg);

	return ret < 0 ? ret : 0;
}

static void ts_set_iface_vlan(struct ts_iface *tsif)
{
	const uint16_t prim_flags = BRIDGE_VLAN_INFO_PVID | BRIDGE_VLAN_INFO_UNTAGGED;
	struct ifinfomsg ifi = { .ifi_family = PF_BRIDGE };
	struct bridge_vlan_info add_vinfo[MAX_VIDS];
	struct bridge_vlan_info del_vinfo[MAX_VIDS];
	int n_del, n_add;
	int i, ret;
	bool has_vid;

	ifi.ifi_index = tsif->ifi_index;
	if (ifi.ifi_index <= 0) {
		warn("Wrong index %d for %s\n", tsif->ifi_index, tsif->name);
		return;
	}

	n_del = 0;
	n_add = 0;

	if (tsif->vid) {
		/* APs - fronthaul/backhaul bss */
		for (i = 0; i < MAX_VIDS; i++) {
			if (tsif->cur_vinfo[i].vid == 0)
				break;

			/* First check VID */
			if (tsif->cur_vinfo[i].vid != tsif->vid &&
			    tsif->cur_vinfo[i].vid != ts_primary_vid) {
				del_vinfo[n_del++] = tsif->cur_vinfo[i];
				continue;
			}

			/* Next check flags */
			if (tsif->cur_vinfo[i].vid == tsif->vid &&
			    tsif->cur_vinfo[i].flags != prim_flags) {
				del_vinfo[n_del++] = tsif->cur_vinfo[i];
				continue;
			}

			/* Check primary vid flags */
			if (tsif->vid != ts_primary_vid &&
			    tsif->cur_vinfo[i].vid == ts_primary_vid &&
			    tsif->cur_vinfo[i].flags != BRIDGE_VLAN_INFO_UNTAGGED) {
				del_vinfo[n_del++] = tsif->cur_vinfo[i];
				continue;
			}
		}

		has_vid = false;
		/* Add configured vid */
		for (i = 0; i < MAX_VIDS; i++) {
			if (tsif->cur_vinfo[i].vid == 0)
				break;

			if (tsif->cur_vinfo[i].vid == tsif->vid &&
			    tsif->cur_vinfo[i].flags == prim_flags) {
				has_vid = true;
				break;
			}
		}

		if (!has_vid) {
			add_vinfo[n_add].vid = tsif->vid;
			add_vinfo[n_add].flags = prim_flags;
			n_add++;
		}

		has_vid = false;
		/* Add additional primary vid untagged */
		if (tsif->vid != ts_primary_vid) {
			for (i = 0; i < MAX_VIDS; i++) {
				if (tsif->cur_vinfo[i].vid == 0)
					break;

				if (tsif->cur_vinfo[i].vid == ts_primary_vid &&
				    tsif->cur_vinfo[i].flags == BRIDGE_VLAN_INFO_UNTAGGED) {
					has_vid = true;
					break;
				}
			}

			if (!has_vid) {
				add_vinfo[n_add].vid = ts_primary_vid;
				add_vinfo[n_add].flags = BRIDGE_VLAN_INFO_UNTAGGED;
				n_add++;
			}
		}

	} else {
		/* APVLAN (wdsX.X.X or wlanX.staY) and backhaul STA - trunk between bridges/devices */
		for (i = 0; i < MAX_VIDS; i++) {
			if (tsif->cur_vinfo[i].vid == 0)
				break;

			if (!ts_is_our_vid(tsif->cur_vinfo[i].vid))
				del_vinfo[n_del++] = tsif->cur_vinfo[i];
		}

		for (i = 0; i < ts_num_vids; i++) {
			add_vinfo[i].vid = ts_all_vids[i];
#if 0
			/* if we want primary vid to be send untagged
			 * similar like for eth interfaces, but this
			 * is not MultiAP spec compliant.
			 */
			if (add_vinfo[i].vid == ts_primary_vid)
				add_vinfo[i].flags = prim_flags;
			else
#endif
			add_vinfo[i].flags = 0;
		}
		n_add = i;
	}

	if (n_add == 0 && n_del == 0)
		return;

	dbg("add %d del %d vids for %s iface %s %d\n", n_add, n_del,
	    tsif->vid ? "fronthaul" : "backhaul", tsif->name, tsif->ifi_index);


	if (n_del > 0) {
		ret = send_bridge_vlan_info(RTM_DELLINK, &ifi, del_vinfo, n_del);
		if (ret < 0)
			warn("Failed to remove vlan info %s\n", tsif->name);
	}

	if (n_add > 0) {
		ret = send_bridge_vlan_info(RTM_SETLINK, &ifi, add_vinfo, n_add);
		if (ret < 0)
			warn("Failed to add vlan info for %s\n", tsif->name);
	}
}

static void ts_set_system(void)
{
	char buf[16] = {0};
	uint16_t vid;
	int i;

	if (ts_primary_vid == 0)
		return;

	for (i = 0; i < ts_num_vids; i++) {
		vid = ts_all_vids[i];
		warn("/lib/wifi/multiap ts create %u\n",vid);
		Cmd(buf, sizeof(buf), "/lib/wifi/multiap ts create %u", vid);
	}

	warn("/lib/wifi/multiap ts reload\n");
	Cmd(buf, sizeof(buf), "/lib/wifi/multiap ts reload");
}

static void ts_check_ifi(struct ifinfomsg *ifi, const char *ifname)
{
	struct ts_iface *tsif;


	if (ts_primary_vid == 0)
		return;

/*
	if (ifi->ifi_family == AF_BRIDGE)
		return;
*/

	if (!(ifi->ifi_flags & IFF_UP))
		return;

	tsif = ts_find_iface(ifname);
	if (tsif) {
		ts_check_vlan_settings = true;
		dbg("found %s tsif %p ifi_index %d\n", ifname, tsif, tsif->ifi_index);
		return;
	}

	if (!strstr(ifname, "wds") && !strstr(ifname, "sta"))
		return;

	tsif = ts_configure_iface(ifname, 0);
	if (tsif) {
		tsif->ifi_index = ifi->ifi_index;
		ts_check_vlan_settings = true;
		dbg("added %s tsif %p ifi_index %d\n", ifname, tsif, tsif->ifi_index);
	} else {
		warn("Fail to add %s interface to ts array\n", ifname);
	}
}

int if_updown(const char *ifname, bool up)
{
	int fd;
	struct ifreq ifr;
	short flags;

	if ((fd = socket(AF_INET, SOCK_STREAM, 0)) < 0) {
		return -1;
	}

	memset(&ifr, 0, sizeof(ifr));
	strncpy(ifr.ifr_name, ifname, IFNAMSIZ);

	if (ioctl(fd, SIOCGIFFLAGS, &ifr) != 0) {
		close(fd);
		return -1;
	}

	flags = ifr.ifr_flags;

	if (up && !(flags & IFF_UP))
		ifr.ifr_flags |= IFF_UP;

	if (!up && (flags & IFF_UP))
		ifr.ifr_flags &= ~IFF_UP;

	if ((flags != ifr.ifr_flags) && (0 != ioctl(fd, SIOCSIFFLAGS, &ifr))) {
		close(fd);
		return -1;
	}

	close(fd);
	return 0;
}

static int nl_main_cb(struct nl_msg *msg, void *arg)
{
	struct nlmsghdr *nlh = nlmsg_hdr(msg);
	struct agent *a = (struct agent *) arg;
	const char *bridge = a->cfg.al_bridge;
	bool brcm_setup = a->cfg.brcm_setup;

	switch (nlh->nlmsg_type) {
	case RTM_GETLINK:
	case RTM_NEWLINK:
	//case RTM_DELLINK:
		{
			struct ifinfomsg *ifi;
			char ifname[16] = {0};
			int ret;

			ifi = NLMSG_DATA(nlh);

			if (if_indextoname(ifi->ifi_index, ifname) == NULL)
				break;

			fprintf(stderr, "ifname = %s  ifindex = %d family = %u status = %s  action = %s\n",
			       ifname,
			       ifi->ifi_index,
			       ifi->ifi_family,
			       !!(ifi->ifi_flags & IFF_UP) ? "up" : "down",
			       nlh->nlmsg_type == RTM_GETLINK ? "RTM_GETLINK" :
			       nlh->nlmsg_type == RTM_NEWLINK ? "RTM_NEWLINK" :
			       "RTM_DELLINK");

/*
			if (!(ifi->ifi_flags & IFF_UP))
				break;
*/
			ts_check_ifi(ifi, ifname);

			if (!brcm_setup)
				break;

			if (!strstr(ifname, "wds"))
				break;

			if (if_isbridge_interface(ifname)) {
				int i, num, max = 32;
				char if_list[max][16];
				bool found = false;

				if (br_get_iflist(bridge, &num, if_list)) {
					printf("Failed to check bridge affiliation\n");
					break;
				}
				max = (num < max) ? num : max;
				for (i = 0; i < max; i++) {
					if (!strncmp(ifname, if_list[i], 15)) {
						printf("Interface %s is already a part of bridge\n",
						       ifname);
						found = true;
						break;
					}
				}
				if (found)
					break;
			}

			printf("Adding interface %s to bridge %s!\n", ifname, bridge);

			/* add wds iface to bridge */
			ret = br_addif(bridge, ifname);
			if (!ret)
				printf("Successfully added interface %s to bridge %s\n",
				       ifname, bridge);

			/* bring up wds interface */
			//ret = if_setflags(ifname, IFF_UP);
			ret = if_updown(ifname, true);
			if (!ret)
				printf("Successfully brought up interface %s\n",
						ifname);
			break;
		}
	default:
		break;

	}

	return 0;
}

static inline bool is_vid_valid(unsigned int vid)
{
#if 0
	dbg("%s: vid %u\n", __func__,  vid);

	if (vid > TS_VID_INVALID)
		abort();
#endif
	return (vid < TS_VID_INVALID) && (vid > 0);
}

static void ts_configure(struct agent *a)
{
	struct netif_fh *fh;
	struct netif_bk *bk;
	struct ts_iface *tsif;
	uint16_t vid;
	char buf[16];

	/* TODO remove interfaces */

	pthread_mutex_lock(&a->nl_mutex);

	ts_primary_vid = a->cfg.pcfg->pvid;

	if (is_vid_valid(a->cfg.pcfg->pvid)) {
		snprintf(buf, sizeof(buf), "%u", ts_primary_vid);
		setenv("PRIMARY_VID", buf, 1);
	} else {
		memset(ts_all_vids, 0, sizeof(ts_all_vids));
		ts_num_vids = 0;
		unsetenv("PRIMARY_VID");
	}

	/* Fronthaul/Backhaul APs */
	list_for_each_entry(fh, &a->fhlist, list) {
		dbg("FH/BH AP name %s multi_ap %d\n", fh->name, fh->cfg->multi_ap);
		vid = fh->cfg->vid;
		if (!is_vid_valid(vid))
			vid = 1;

		ts_configure_iface(fh->name, vid);
	}

	/* Backhaul stations */
	list_for_each_entry(bk, &a->bklist, list) {
		dbg("BK name %s\n", bk->name);

		tsif = ts_configure_iface(bk->name, 0);
		if (tsif)
			tsif->is_bsta = true;
	}

	pthread_mutex_unlock(&a->nl_mutex);
}

int init_bridge_nl_sock(void)
{
	nl_bridge_sk = nl_socket_alloc();
	if (!nl_bridge_sk)
		return -1;

	nl_bridge_cb = nl_cb_alloc(NL_CB_DEFAULT);
	if (!nl_bridge_cb) {
		nl_socket_free(nl_bridge_sk);
		return -1;
	}

	nl_connect(nl_bridge_sk, NETLINK_ROUTE);
	nl_socket_add_memberships(nl_bridge_sk, RTNLGRP_LINK, 0);

	nl_cb_set(nl_bridge_cb, NL_CB_VALID, NL_CB_CUSTOM, ts_nl_get_cb, NULL);
	nl_cb_set(nl_bridge_cb, NL_CB_FINISH, NL_CB_CUSTOM, ts_nl_finish_cb, NULL);
	nl_cb_set(nl_bridge_cb, NL_CB_ACK, NL_CB_CUSTOM, ts_nl_finish_cb, NULL);
	nl_cb_err(nl_bridge_cb, NL_CB_CUSTOM, ts_nl_error_cb, NULL);

	return 0;
}

int init_main_nl_sock(struct agent *agent)
{
	nl_main_sk = nl_socket_alloc();
	if (!nl_main_sk)
		return -1;

	nl_socket_disable_seq_check(nl_main_sk);
	nl_socket_modify_cb(nl_main_sk, NL_CB_VALID, NL_CB_CUSTOM, nl_main_cb, agent);

	nl_connect(nl_main_sk, NETLINK_ROUTE);
	nl_socket_add_memberships(nl_main_sk, RTNLGRP_LINK, 0);
	nl_rtgen_request(nl_main_sk, RTM_GETLINK, AF_UNSPEC, NLM_F_DUMP);

	return 0;
}

int nl_loop(void *arg)
{
	struct agent *agent = arg;
	char *bridge = agent->cfg.al_bridge;
	struct pollfd fds[1];
	int ret, i;

	setenv("AL_BRIDGE", bridge, 1);

	ret = init_bridge_nl_sock();
	if (ret)
		return -1;

	ret = init_main_nl_sock(agent);
	if (ret)
		return -1;

	fds[0].fd = nl_socket_get_fd(nl_main_sk);
	fds[0].events = POLLIN;

	for (;;) {
		ts_check_vlan_settings = false;

		dbg("Start polling ...\n");

		ret = poll(fds, 1, -1);

		dbg("End polling with %d ret %08x revents\n", ret , fds[0].revents);

		if (fds[0].revents) {
			ret = nl_recvmsgs_default(nl_main_sk);
			fprintf(stderr, "nl_recvmsgs_default ret %d\n", ret);
		}

		if (ts_need_reconfig) {
			ts_configure(agent);
			ts_need_reconfig = false;

			ts_set_system();
		}

		if (ts_check_vlan_settings) {
			dbg("Reading bridge %s vlan config\n", bridge);
			for (i = 0; i < TS_IFACE_MAX_NUM; i++) {
				int j;

				for (j = 0; j < MAX_VIDS; j++) {
					ts_iface_array[i].cur_vinfo[j].vid = 0;
				}

				ts_iface_array[i].ifi_index = 0;
			}
			ts_read_bridge_vlan(bridge);

			for (i = 0; i < TS_IFACE_MAX_NUM; i++) {
				struct ts_iface *tsif = &ts_iface_array[i];

				if (tsif->name[0] == '\0')
					break;

				dbg("checking vlan settings for %s tsif %p index %d\n",
				    tsif->name, tsif, tsif->ifi_index);

				/* Add to the bridge if not added already and if not backhaul sta,
				 * bsta bridging is done by dynamically by other entities.
				 */
				if (ts_iface_array[i].ifi_index <= 0 && !tsif->is_bsta) {
					ret = br_addif(bridge, tsif->name);
					dbg("(re)added %s interface to %s with status %d\n",
					    tsif->name, bridge, ret);

					ts_iface_array[i].ifi_index = if_nametoindex(tsif->name);
					dbg("get %s index %d\n", tsif->name, tsif->ifi_index);

				}

				if (ts_iface_array[i].ifi_index <= 0)
					continue;
				ts_set_iface_vlan(&ts_iface_array[i]);
			}
		}
	}

	return 0;
}
